==> conc-1.9.4.tar.gz: conc-1.9.4/client/console.c <==
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <termios.h>
#include <errno.h>
#include <libgnome/libgnome.h>
#include <gnet.h>
#include "i18n.h"
#include "config.h"
#include "mplex.h"
#include "message.h"
#include "client-message.h"

#define ESCAPE_CHAR 1

GMainLoop *main_loop;
Mplex mplex;
MessageQueue msg_q;
struct termios orig_tio;
const gchar *console;
GIOChannel *io;
GIOChannel *stdin_io;
GIOChannel *stdout_io;

static void
restore_terminal(void) {
    tcsetattr(0, TCSANOW, &orig_tio);
}

static void
console_detach(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    if (arg->type != MESSAGE_ARG_STRING) {
	g_printerr("malformed message\n");
	exit(EXIT_FAILURE);
    }
    /* FIXME: should we bother to check the name of the console? */
    restore_terminal();
    g_print("\n\nThe session has been stopped because console %s\n"
	    "is no longer available.\n", arg->string->str);
    g_main_loop_quit(main_loop);
}

static gboolean
read_remote(GIOChannel *source,
	GIOCondition condition,
	gpointer data) {
    gchar buf[256];
    gsize bytes;
    GError *e = NULL;

    switch (g_io_channel_read_chars(source, buf, sizeof(buf), &bytes, &e)) {
	case G_IO_STATUS_ERROR:
	    /* FIXME: report error? */
	    return FALSE;
	case G_IO_STATUS_NORMAL:
	    break;
	case G_IO_STATUS_EOF:
	    return FALSE;
	case G_IO_STATUS_AGAIN:
	    return TRUE;
	default:
	    g_assert_not_reached();
    }
    g_io_channel_write_chars(stdout_io, buf, bytes, &bytes, &e);
    /* FIXME: check error */
    return TRUE;
}

static gboolean
read_local(GIOChannel *source,
	GIOCondition condition,
	gpointer data) {
    static gboolean escaped = FALSE;
    gchar buf;
    gsize bytes;
    GError *e = NULL;

    switch (g_io_channel_read_chars(source, &buf, 1, &bytes, &e)) {
	case G_IO_STATUS_ERROR:
	    /* FIXME: report error? */
	    return FALSE;
	case G_IO_STATUS_NORMAL:
	    break;
	case G_IO_STATUS_EOF:
	    return FALSE;
	case G_IO_STATUS_AGAIN:
	    return TRUE;
	default:
	    g_assert_not_reached();
    }
    if (escaped == TRUE) {
	switch (buf) {
	    case 'q':
		g_main_loop_quit(main_loop);
		break;
	    case 'a':
	    case 'A':
		buf = ESCAPE_CHAR;
	    default:
		g_io_channel_write_chars(io, &buf, bytes, &bytes, &e);
		/* FIXME: check error */
	}
	escaped = FALSE;
    } else {
	if (buf == ESCAPE_CHAR) {
	    escaped = TRUE;
	} else {
	    g_io_channel_write_chars(io, &buf, bytes, &bytes, &e);
	    /* FIXME: check error */
	}
    }
    return TRUE;
}

static void
attached(MessageQueue q,
	const MessageArg *arg,
	gpointer data) {
    GError *e = message_arg_error_check(arg);
    struct termios tio;

    if (e != NULL) {
	g_printerr(_("failed to attach to console %s (%s)\n"),
		console, e->message);
	exit(EXIT_FAILURE);
    }
    if (arg->type != MESSAGE_ARG_UINT16) {
	g_printerr(_("bad response from manager\n"));
	exit(EXIT_FAILURE);
    }
    io = mplex_open(mplex, arg->uint16, NULL, &e);
    if (io == NULL) {
	g_printerr(_("failed to open channel to console %s (%s)\n"),
		console, e->message);
	exit(EXIT_FAILURE);
    }
    g_print(_("Connected to console %s. Use ^A-q to quit\n"), console);
    tio = orig_tio;
    cfmakeraw(&tio);
    tio.c_cflag |= HUPCL | CLOCAL | CREAD | CS8;
    tio.c_lflag |= NOFLSH;
    tio.c_lflag &= ~(ICANON | TOSTOP | ECHO | ECHOE | ECHOK | ECHONL |
	    IEXTEN | ISIG);
    tio.c_iflag &= ~(IGNBRK | BRKINT | IGNPAR | PARMRK | INPCK | ISTRIP |
	    INLCR | IGNCR | ICRNL | IXON | IXOFF);
    tio.c_oflag &= ~OPOST;
    tio.c_cc[VMIN] = 0;
    tio.c_cc[VTIME] = 0;
    if (tcsetattr(0, TCSANOW, &tio) == -1) {
	g_printerr(_("cannot set terminal attributes (%s)\n"),
		g_strerror(errno));
	exit(EXIT_FAILURE);
    }
    stdin_io = g_io_channel_unix_new(0);
    if (g_io_channel_get_encoding(stdin_io) != NULL) {
        g_io_channel_set_encoding(stdin_io, NULL, NULL);
    }
    g_io_channel_set_buffered(stdin_io, FALSE);
    stdout_io = g_io_channel_unix_new(1);
    if (g_io_channel_get_encoding(stdout_io) != NULL) {
        g_io_channel_set_encoding(stdout_io, NULL, NULL);
    }
    g_io_channel_set_buffered(stdout_io, FALSE);
    g_io_add_watch(g_io_channel_unix_new(0), G_IO_IN, read_local, NULL);
    g_io_add_watch(io, G_IO_IN, read_remote, NULL);
}

static void
logged_in(MessageQueue q,
	const MessageArg *arg,
	gpointer data) {
    GError *e = message_arg_error_check(arg);
    MessageArg *a;

    if (e != NULL) {
	g_printerr("failed to log in (%s)\n", e->message);
	exit(EXIT_FAILURE);
    }
    a = message_arg_string("console", console);
    if (message_send(msg_q, CLIENT_MESSAGE_ATTACH, a, attached,
		NULL, &e) == FALSE) {
	g_printerr("failed to send message (%s)\n", e->message);
	exit(EXIT_FAILURE);
    }
    message_arg_free(a);
}

static void
connection_closed(Mplex mplex,
	gpointer data) {
    g_printerr("connection closed by manager\n");
    exit(EXIT_FAILURE);
}

int
main(int argc, char *argv[]) {
    GnomeProgram *program;
    poptContext pc;
    gint debug = 0;
    gchar *host = DEFAULT_MANAGER;
    const gchar *user = g_get_user_name();
    gint port = CONC_PORT_MGR;
    struct poptOption options[] = {
	{ "debug", 'd', POPT_ARG_NONE, &debug, 1, N_("Enter debug mode") },
	{ "host", 'h', POPT_ARG_STRING, &host, 0,
	    N_("Host running manager"), N_("HOST") },
	{ "user", 'u', POPT_ARG_STRING, &user, 0,
	    N_("Conc user name"), N_("USER") },
	POPT_AUTOHELP
	{ NULL, 0, 0, NULL, 0 }
    };
    gchar pass_buf[256];
    gint i;
    GTcpSocket *sock;
    GError *e = NULL;
    GIOChannel *io;
    struct termios tio;
    GList *l = NULL;
    MessageArg *arg;

    program = gnome_program_init("console", PACKAGE_VERSION,
	LIBGNOME_MODULE, argc, argv,
	GNOME_PARAM_POPT_TABLE, options,
	GNOME_PARAM_NONE);
    main_loop = g_main_loop_new(NULL, FALSE);
    g_object_get(program, GNOME_PARAM_POPT_CONTEXT, &pc, NULL);
    console = poptGetArg(pc);
    if (console == NULL) {
	g_printerr("error: you must specify a console to connect to\n");
	exit(EXIT_FAILURE);
    }
    if (isatty(0) == 0) {
	g_printerr("%s must be run from a terminal\n",
		gnome_program_get_human_readable_name(program));
	exit(EXIT_FAILURE);
    }
    if (tcgetattr(0, &orig_tio) == -1) {
	g_printerr("cannot get terminal attributes (%s)\n",
		g_strerror(errno));
	exit(EXIT_FAILURE);
    }
    g_atexit(restore_terminal);
    tio = orig_tio;
    tio.c_lflag &= ~(ECHO | ECHOE);
    if (tcsetattr(0, TCSANOW, &tio) == -1) {
	g_printerr("cannot set terminal attributes (%s)\n",
		g_strerror(errno));
	exit(EXIT_FAILURE);
    }
    g_print("Password: ");
    pass_buf[0] = '\0';
    fgets(pass_buf, 256, stdin);
    g_print("\n");
    if (tcsetattr(0, TCSANOW, &tio) == -1) {
	g_printerr("cannot set terminal attributes (%s)\n",
		g_strerror(errno));
	exit(EXIT_FAILURE);
    }
    tio = orig_tio;
    for (i = 0; pass_buf[i] != '\0'; i++) {
	if (pass_buf[i] == '\n') {
	    pass_buf[i] = '\0';
	    break;
	}
    }
    sock = gnet_tcp_socket_connect(host, port);
    if (sock == NULL) {
	g_printerr("failed to connect to manager on %s\n", host);
	exit(EXIT_FAILURE);
    }
    mplex = mplex_new(gnet_tcp_socket_get_io_channel(sock),
	    connection_closed, NULL, &e);
    if (e != NULL) {
	g_printerr("failed to start mplex (%s)\n", e->message);
	exit(EXIT_FAILURE);
    }
    io = mplex_open(mplex, CONC_MSG_CHANNEL, NULL, &e);
    if (e != NULL) {
	g_printerr("failed to open channel for message queue (%s)\n",
		e->message);
	exit(EXIT_FAILURE);
    }
    msg_q = message_queue_new(io, &e);
    if (e != NULL) {
	g_printerr("failed to open message queue (%s)\n", e->message);
	exit(EXIT_FAILURE);
    }
    /* register handlers */
    if (message_handler_register(msg_q, CLIENT_MESSAGE_DETACH,
		console_detach, NULL, &e) == FALSE) {
	g_printerr("failed to register message handler (%s)\n", e->message);
	exit(EXIT_FAILURE);
    }
    message_queue_start(msg_q);
    mplex_start(mplex);
    l = g_list_append(l, message_arg_string("user", user));
    l = g_list_append(l, message_arg_string("password", pass_buf));
    arg = message_arg_list("", l);
    if (message_send(msg_q, CLIENT_MESSAGE_LOGIN, arg, logged_in,
		NULL, &e) == FALSE) {
	g_printerr("failed to send message (%s)\n", e->message);
	exit(EXIT_FAILURE);
    }
    message_arg_free(arg);
    for (i = 0; pass_buf[i] != '\0'; i++) {
	pass_buf[i] = 'X';
    }
    g_main_loop_run(main_loop);
    exit(EXIT_SUCCESS);
}

==> conc-2.0.1.tar.gz: conc-2.0.1/client/console.c <==
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <termios.h>
#include <errno.h>
#include <libgnome/libgnome.h>
#include <gnet.h>
#include "i18n.h"
#include "config.h"
#include "mplex.h"
#include "message.h"
#include "client-message.h"

#define ESCAPE_CHAR 1

GMainLoop *main_loop;
Mplex mplex;
MessageQueue msg_q;
struct termios orig_tio;
const gchar *console;
GIOChannel *io;
GIOChannel *stdin_io;
GIOChannel *stdout_io;

static void
restore_terminal(void) {
    tcsetattr(0, TCSANOW, &orig_tio);
}

static void
console_detach(MessageQueue q,
	const gchar *name,
	MessageID id,
	const MessageArg *arg,
	gpointer data) {
    if (arg->type != MESSAGE_ARG_STRING) {
	g_printerr("malformed message\n");
	exit(EXIT_FAILURE);
    }
    /* FIXME: should we bother to check the name of the console? */
    restore_terminal();
    g_print("\n\nThe session has been stopped because console %s\n"
	    "is no longer available.\n", arg->string->str);
    g_main_loop_quit(main_loop);
}

static gboolean
read_remote(GIOChannel *source,
	GIOCondition condition,
	gpointer data) {
    gchar buf[256];
    gsize bytes;
    GError *e = NULL;

    switch (g_io_channel_read_chars(source, buf, sizeof(buf), &bytes, &e)) {
	case G_IO_STATUS_ERROR:
	    /* FIXME: report error? */
	    return FALSE;
	case G_IO_STATUS_NORMAL:
	    break;
	case G_IO_STATUS_EOF:
	    return FALSE;
	case G_IO_STATUS_AGAIN:
	    return TRUE;
	default:
	    g_assert_not_reached();
    }
    g_io_channel_write_chars(stdout_io, buf, bytes, &bytes, &e);
    /* FIXME: check error */
    return TRUE;
}

static gboolean
read_local(GIOChannel *source,
	GIOCondition condition,
	gpointer data) {
    static gboolean escaped = FALSE;
    gchar buf;
    gsize bytes;
    GError *e = NULL;

    switch (g_io_channel_read_chars(source, &buf, 1, &bytes, &e)) {
	case G_IO_STATUS_ERROR:
	    /* FIXME: report error? */
	    return FALSE;
	case G_IO_STATUS_NORMAL:
	    break;
	case G_IO_STATUS_EOF:
	    return FALSE;
	case G_IO_STATUS_AGAIN:
	    return TRUE;
	default:
	    g_assert_not_reached();
    }
    if (escaped == TRUE) {
	switch (buf) {
	    case 'q':
		g_main_loop_quit(main_loop);
		break;
	    case 'a':
	    case 'A':
		buf = ESCAPE_CHAR;
	    default:
		g_io_channel_write_chars(io, &buf, bytes, &bytes, &e);
		/* FIXME: check error */
	}
	escaped = FALSE;
    } else {
	if (buf == ESCAPE_CHAR) {
	    escaped = TRUE;
	} else {
	    g_io_channel_write_chars(io, &buf, bytes, &bytes, &e);
	    /* FIXME: check error */
	}
    }
    return TRUE;
}

static void
attached(MessageQueue q,
	const MessageArg *arg,
	gpointer data) {
    GError *e = message_arg_error_check(arg);
    struct termios tio;

    if (e != NULL) {
	g_printerr(_("failed to attach to console %s (%s)\n"),
		console, e->message);
	exit(EXIT_FAILURE);
    }
    if (arg->type != MESSAGE_ARG_UINT16) {
	g_printerr(_("bad response from manager\n"));
	exit(EXIT_FAILURE);
    }
    io = mplex_open(mplex, arg->uint16, NULL, &e);
    if (io == NULL) {
	g_printerr(_("failed to open channel to console %s (%s)\n"),
		console, e->message);
	exit(EXIT_FAILURE);
    }
    g_print(_("Connected to console %s. Use ^A-q to quit\n"), console);
    tio = orig_tio;
    cfmakeraw(&tio);
    tio.c_cflag |= HUPCL | CLOCAL | CREAD | CS8;
    tio.c_lflag |= NOFLSH;
    tio.c_lflag &= ~(ICANON | TOSTOP | ECHO | ECHOE | ECHOK | ECHONL |
	    IEXTEN | ISIG);
    tio.c_iflag &= ~(IGNBRK | BRKINT | IGNPAR | PARMRK | INPCK | ISTRIP |
	    INLCR | IGNCR | ICRNL | IXON | IXOFF);
    tio.c_oflag &= ~OPOST;
    tio.c_cc[VMIN] = 0;
    tio.c_cc[VTIME] = 0;
    if (tcsetattr(0, TCSANOW, &tio) == -1) {
	g_printerr(_("cannot set terminal attributes (%s)\n"),
		g_strerror(errno));
	exit(EXIT_FAILURE);
    }
    stdin_io = g_io_channel_unix_new(0);
    if (g_io_channel_get_encoding(stdin_io) != NULL) {
        g_io_channel_set_encoding(stdin_io, NULL, NULL);
    }
    g_io_channel_set_buffered(stdin_io, FALSE);
    stdout_io = g_io_channel_unix_new(1);
    if (g_io_channel_get_encoding(stdout_io) != NULL) {
        g_io_channel_set_encoding(stdout_io, NULL, NULL);
    }
    g_io_channel_set_buffered(stdout_io, FALSE);
    g_io_add_watch(g_io_channel_unix_new(0), G_IO_IN, read_local, NULL);
    g_io_add_watch(io, G_IO_IN, read_remote, NULL);
}

static void
logged_in(MessageQueue q,
	const MessageArg *arg,
	gpointer data) {
    GError *e = message_arg_error_check(arg);
    MessageArg *a;

    if (e != NULL) {
	g_printerr("failed to log in (%s)\n", e->message);
	exit(EXIT_FAILURE);
    }
    a = message_arg_string("console", console);
    if (message_send(msg_q, CLIENT_MESSAGE_ATTACH, a, attached,
		NULL, &e) == FALSE) {
	g_printerr("failed to send message (%s)\n", e->message);
	exit(EXIT_FAILURE);
    }
    message_arg_free(a);
}

static void
connection_closed(Mplex mplex,
	gpointer data) {
    g_printerr("connection closed by manager\n");
    exit(EXIT_FAILURE);
}

int
main(int argc, char *argv[]) {
    GnomeProgram *program;
    gint debug = 0;
    gchar *host = DEFAULT_MANAGER;
    const gchar *user = g_get_user_name();
    gint port = CONC_PORT_MGR;
    gchar **remaining_args = NULL;
    GOptionEntry options[] = {
	{ "debug", 'd', 0, G_OPTION_ARG_NONE, &debug,
	    N_("Enter debug mode"), NULL },
	{ "host", 'h', 0, G_OPTION_ARG_STRING, &host,
	    N_("Host running manager"), N_("HOST") },
	{ "user", 'u', 0, G_OPTION_ARG_STRING, &user,
	    N_("Conc user name"), N_("USER") },
	{ G_OPTION_REMAINING, 0, 0, G_OPTION_ARG_FILENAME_ARRAY,
	    &remaining_args, "console to connect to", "CONSOLE" },
	{ NULL }
    };
    GOptionContext *option_context;
    gchar pass_buf[256];
    gint i;
    GTcpSocket *sock;
    GError *e = NULL;
    GIOChannel *io;
    struct termios tio;
    GList *l = NULL;
    MessageArg *arg;


    option_context = g_option_context_new(NULL);
    g_option_context_add_main_entries(option_context, options,
	GETTEXT_PACKAGE);
    program = gnome_program_init("console", PACKAGE_VERSION,
	LIBGNOME_MODULE, argc, argv,
	GNOME_PARAM_GOPTION_CONTEXT, option_context,
	GNOME_PARAM_NONE);
    main_loop = g_main_loop_new(NULL, FALSE);
    if (remaining_args == NULL) {
	g_printerr("error: you must specify a console to connect to\n");
	exit(EXIT_FAILURE);
    }
    console = remaining_args[0];
    if (isatty(0) == 0) {
	g_printerr("%s must be run from a terminal\n",
		gnome_program_get_human_readable_name(program));
	exit(EXIT_FAILURE);
    }
    if (tcgetattr(0, &orig_tio) == -1) {
	g_printerr("cannot get terminal attributes (%s)\n",
		g_strerror(errno));
	exit(EXIT_FAILURE);
    }
    g_atexit(restore_terminal);
    tio = orig_tio;
    tio.c_lflag &= ~(ECHO | ECHOE);
    if (tcsetattr(0, TCSANOW, &tio) == -1) {
	g_printerr("cannot set terminal attributes (%s)\n",
		g_strerror(errno));
	exit(EXIT_FAILURE);
    }
    g_print("Password: ");
    pass_buf[0] = '\0';
    fgets(pass_buf, 256, stdin);
    g_print("\n");
    if (tcsetattr(0, TCSANOW, &tio) == -1) {
	g_printerr("cannot set terminal attributes (%s)\n",
		g_strerror(errno));
	exit(EXIT_FAILURE);
    }
    tio = orig_tio;
    for (i = 0; pass_buf[i] != '\0'; i++) {
	if (pass_buf[i] == '\n') {
	    pass_buf[i] = '\0';
	    break;
	}
    }
    sock = gnet_tcp_socket_connect(host, port);
    if (sock == NULL) {
	g_printerr("failed to connect to manager on %s\n", host);
	exit(EXIT_FAILURE);
    }
    mplex = mplex_new(gnet_tcp_socket_get_io_channel(sock),
	    connection_closed, NULL, &e);
    if (e != NULL) {
	g_printerr("failed to start mplex (%s)\n", e->message);
	exit(EXIT_FAILURE);
    }
    io = mplex_open(mplex, CONC_MSG_CHANNEL, NULL, &e);
    if (e != NULL) {
	g_printerr("failed to open channel for message queue (%s)\n",
		e->message);
	exit(EXIT_FAILURE);
    }
    msg_q = message_queue_new(io, &e);
    if (e != NULL) {
	g_printerr("failed to open message queue (%s)\n", e->message);
	exit(EXIT_FAILURE);
    }
    /* register handlers */
    if (message_handler_register(msg_q, CLIENT_MESSAGE_DETACH,
		console_detach, NULL, &e) == FALSE) {
	g_printerr("failed to register message handler (%s)\n", e->message);
	exit(EXIT_FAILURE);
    }
    message_queue_start(msg_q);
    mplex_start(mplex);
    l = g_list_append(l, message_arg_string("user", user));
    l = g_list_append(l, message_arg_string("password", pass_buf));
    arg = message_arg_list("", l);
    if (message_send(msg_q, CLIENT_MESSAGE_LOGIN, arg, logged_in,
		NULL, &e) == FALSE) {
	g_printerr("failed to send message (%s)\n", e->message);
	exit(EXIT_FAILURE);
    }
    message_arg_free(arg);
    for (i = 0; pass_buf[i] != '\0'; i++) {
	pass_buf[i] = 'X';
    }
    g_main_loop_run(main_loop);
    exit(EXIT_SUCCESS);
}
